from enum import Enum


class Space(Enum):
    K8 = "8k"
    K32 = "32k"
    K59 = "64k"

    def __str__(self):
        return self.value


class TopoName(Enum):
    T1T2 = "t1t2"
    MARGULIES_GRADIENT = "gradient"
    ANT_POST_GRADIENT = "coord"
    MEDIAL_WALL = "medial_wall"

    def __str__(self):
        return self.value


class TemplateName(Enum):
    SCHAEFER_200_7 = "sh2007"
    SCHAEFER_200_17 = "sh20017"
    COLE_360 = "cole"
    WANG = "wang"

    def __str__(self):
        return self.value



import numpy as np
from scipy import signal
from statsmodels.tsa import stattools


def lempel_ziv_complexity(sequence):
    sub_strings = set()

    ind = 0
    inc = 1
    while True:
        if ind + inc > len(sequence):
            break
        sub_str = sequence[ind : ind + inc]
        if sub_str in sub_strings:
            inc += 1
        else:
            sub_strings.add(sub_str)
            ind += inc
            inc = 1
    return len(sub_strings)


def calc_a_acf(ts, n_lag=None, fast=True):
    if not n_lag:
        n_lag = len(ts)
    return stattools.acf(ts, nlags=n_lag, qstat=False, alpha=None, fft=fast)


def calc_a_acw(ts, n_lag=None, fast=True, is_acf=False):
    acf = ts if is_acf else calc_a_acf(ts, n_lag, fast)
    return np.argmax(acf < 0.5)


def calc_a_acz(ts, n_lag=None, fast=True, is_acf=False):
    acf = ts if is_acf else calc_a_acf(ts, n_lag, fast)
    return np.argmax(acf <= 0)


def calc_a_acmi(ts, which, n_lag=None, fast=True, is_acf=False):
    acf = ts if is_acf else calc_a_acf(ts, n_lag, fast)
    return signal.argrelextrema(acf, np.less)[0][which - 1]


def calc_lzc_norm_factor(ts):
    """
    The default way of calculating LZC normalization factor for a time series
    :param ts: a time series
    :return: normalization factor
    """
    return len(ts) / np.log2(len(ts))


def calc_lzc(ts, norm_factor=None):
    """
    Calculates lempel-ziv complexity of a single time series.
    :param ts: a time-series: nx1
    :param norm_factor: the normalization factor. If none, the default value will be calculated: n / ln(n)
    :return: the lempel-ziv complexity
    """
    if not norm_factor:
        norm_factor = calc_lzc_norm_factor(ts)
    bin_ts = np.char.mod('%i', ts >= np.median(ts))
    return lempel_ziv_complexity("".join(bin_ts)) / norm_factor


def calc_mf(freq, psd):
    """
    Calculates median frequency of different power spectral densities.
    n is number of PSDs and m in number of frequencies.
    :param freq: the mx1 numpy array of frequencies.
    :param psd: the nxm numpy matrix of power.
    :return: nx1 array of median frequencies.
    """
    cum_sum = np.cumsum(psd, axis=1)
    return freq[np.argmax(cum_sum >= cum_sum[:, -1].reshape(-1, 1) / 2, axis=1)]



import itertools
import pandas as pd
import numpy as np
from matplotlib.cbook import boxplot_stats
import pandas_flavor as pf
from neuro_helper.entity import Space, TemplateName
from neuro_helper.statistics import percent_change, icc


@pf.register_dataframe_method
def add_net_meta(df, labels):
    meta = pd.Series(index=df.index, name="net_meta")
    for label, nets in labels.items():
        meta.loc[df.network.isin(nets)] = label
    df["net_meta"] = meta
    return df


@pf.register_dataframe_method
def convert_column(df, **col_dict):
    new_df = df.copy()
    for col_name, func in col_dict.items():
        new_df[col_name] = func(new_df[col_name])
    return new_df


@pf.register_dataframe_method
def and_filter(df, drop_single=True, **kwargs):
    filt = True
    keys = []
    for key, value in kwargs.items():
        negate = False
        if key.startswith("NOT"):
            negate = True
            key = key.replace("NOT", "")

        keys.append(key)
        if type(value) in [list, tuple, np.ndarray]:
            this_filt = df[key].isin(value)
        else:
            this_filt = df[key] == value

        filt &= ~this_filt if negate else this_filt

    new_df = df[filt]
    if drop_single:
        return new_df.drop([c for c in keys if len(new_df[c].unique()) <= 1], 1)
    else:
        return new_df


@pf.register_dataframe_method
def get_outlier_bounds(df, of):
    if isinstance(of, str):
        of = [of, ]

    out = []
    for col in of:
        stat = boxplot_stats(df[col])[0]
        out.append((stat["whislo"], stat["whishi"]))

    return out[0] if len(out) == 1 else out


@pf.register_dataframe_method
def avg_over_net(df):
    return df.groupby(list(df.columns.drop(["region", "metric"]))).mean().reset_index()


@pf.register_dataframe_method
def normalize(x, columns, new_min=0, new_max=1):
    if isinstance(columns, str):
        columns = [columns, ]
    df = x.copy()
    for on in columns:
        df[on] = normalize_series(df[on], new_min, new_max)
    return df


@pf.register_dataframe_method
def add_topo(df, template_name: TemplateName, space: Space, *args):
    from neuro_helper.template import get_topo_dataframe
    new_df = df
    has_net = "network" in df.columns
    for arg in args:
        topo = get_topo_dataframe(arg, template_name, space)
        if has_net:
            new_df = pd.merge(new_df, topo, on=["region", "network"])
        else:
            topo = topo.drop("network", 1)
            new_df = pd.merge(new_df, topo, on=["region"])

    return new_df


@pf.register_dataframe_method
def add_median_lh(x, calc_med_col, values=("L", "H")):
    med = x[calc_med_col].median()
    return add_split_label(x, calc_med_col, calc_med_col, (values, med))


@pf.register_dataframe_method
def add_split_label(df, on, based, criteria):
    x = df.copy()
    if callable(criteria):
        labels, borders = criteria(x[based])
    else:
        labels, borders = criteria

    if np.isscalar(borders):
        borders = [borders, ]

    if len(labels) != len(borders) + 1:
        raise ValueError("labels should be one more than borders")

    new_col_name = f"{on}_split"
    on_splitted = pd.Series(index=x.index, name=new_col_name, data=pd.NA)

    borders.append(borders[-1])
    for index, (label, border) in enumerate(zip(labels, borders)):
        if index == 0:
            filt = x[based] < border
        elif index == len(labels) - 1:
            filt = x[based] >= border
        else:
            filt = (x[based] < border) & (x[based] >= borders[index - 1])
        on_splitted.loc[filt] = label

    if on_splitted.isna().any():
        raise ValueError(f"criteria does not cover the whole {on} bound")

    x[new_col_name] = on_splitted
    return x


@pf.register_dataframe_method
def remove_outliers(x, of):
    stat = boxplot_stats(x[of])[0]
    low, high = stat["whislo"], stat["whishi"]
    return x.loc[(x[of] > low) & (x[of] < high)]


@pf.register_dataframe_method
def calc_paired_diff(x, diff_func=lambda left, right: abs(left - right), repeat=True):
    """
    calculates the 2-by-2 difference on one single column.
    :param x: a dataframe with only two columns.
    First column is the label for the second column. The second one is the metric
    :param diff_func: the difference function. default is L1 norm
    :param repeat: if True return all combinations with repeteation (product), otherwise only unique combinations
    :return: a dataframe with 3 columns. Left items, Right items and the difference between left and right
    """
    diff = pd.DataFrame(columns=("left", "right", "difference"))
    iterator = itertools.product(range(len(x)), repeat=2) if repeat else itertools.combinations(range(len(x)), 2)
    index = 0
    for li, ri in iterator:
        if li == ri:
            continue
        diff.loc[index, :] = x.iloc[li, 0], x.iloc[ri, 0], diff_func(x.iloc[li, 1], x.iloc[ri, 1])
        index += 1

    diff.difference = diff.difference.astype(np.float)
    return diff


@pf.register_dataframe_method
def calc_pchange(x):
    a = x[x.task == "Rest"]
    if not a.shape[0] == 1:
        raise Exception("Rest does not exist for %s" % x)
    rest_val = a.metric.item()
    output = []
    for task in x.task:
        if task == "Rest":
            continue
        output.append([task, percent_change(rest_val, x[x.task == task].metric.item())])

    df = pd.DataFrame(output, columns=["task", "pchange"])
    df.pchange = df.pchange.astype(float)
    return df


@pf.register_dataframe_method
def calc_icc(x):
    return pd.Series(
        {"icc": icc(x.drop("region", 1).pivot(index='subject', columns='scan', values='metric').values)})


@pf.register_series_method
def normalize_series(x, new_min=0, new_max=1):
    old_min = x.min()
    old_max = x.max()
    old_range = old_max - old_min
    new_range = new_max - new_min
    return (((x - old_min) * new_range) / old_range) + new_min


def concat_dfs(by, on, new_col="cat", **dfs):
    df = pd.DataFrame(columns=on + [by, new_col])
    for label, x in dfs.items():
        temp = pd.DataFrame(data=x, columns=on + [by, new_col])
        temp[new_col] = label
        df = df.append(temp, ignore_index=True, sort=False)

    return df.reset_index(drop=True)



import os
import numpy as np
import cifti
from pandas import DataFrame
from neuro_helper.template import get_template
from neuro_helper.entity import TemplateName, Space
import pandas_flavor as pf


def out_of(name, dir_from_last_part=True):
    file = "outputs"

    name_none, ext = os.path.splitext(name)

    for checker in [".dlabel", ".dtseries", ".shape"]:
        if checker in name_none:
            name_none = name_none.replace(checker, "")
            ext = checker + ext

    if ext in [".png", ".jpg", ".jpeg"]:
        file += "/figures"

    name_splitted = name_none.split(".")
    middle_dirs = os.sep.join(name_splitted[:-1])
    last_dir = name_splitted[-1].split("-")[0] if dir_from_last_part else ""

    file = os.path.join(file, middle_dirs, last_dir, name)

    file_dir = os.path.dirname(file)
    if "*" not in file_dir and not os.path.exists(file_dir):
        os.makedirs(file_dir)
    return file


def find_shared_subjects(find_files, prepare_file_content,
                         template_name: TemplateName, space: Space,
                         tasks, return_indices):
    all_ids = []
    subject_id_dict = {}
    for task in tasks:
        files = find_files(task=task, template_name=template_name, space=space)
        for file in files:
            scan_id, subj_ids = prepare_file_content(np.load(file, allow_pickle=True))
            all_ids += subj_ids
            subject_id_dict["%s-%s" % (task, scan_id)] = np.asarray(subj_ids)

    all_ids.sort()
    all_ids, count = np.unique(all_ids, return_counts=True)
    shared_ids = all_ids[count == len(subject_id_dict.keys())]
    keys = list(subject_id_dict.keys())
    keys.sort()
    shared_indices = np.asarray([np.argwhere(np.isin(subject_id_dict[key], shared_ids)).flatten() for key in keys])
    if return_indices:
        return shared_ids, keys, shared_indices
    else:
        return shared_ids


def generate_long_data(find_files, prepare_file_content, template_name: TemplateName, space: Space):
    mask, unique_networks, networks, regions, _ = get_template(template_name, space)
    long_data = None
    for task in ["Rest", "StoryM", "Motort", "Wrkmem"]:
        files = find_files(task=task, template_name=template_name, space=space)
        for file in files:
            scan_id, subj_ids, metric = prepare_file_content(np.load(file, allow_pickle=True))
            n_subj, n_regions = metric.shape
            task_scan = "%s-%s" % (task, scan_id)
            cols = np.c_[
                np.repeat(task, n_subj * n_regions),
                np.repeat(scan_id, n_subj * n_regions),
                np.repeat(task_scan, n_subj * n_regions),
                np.repeat(subj_ids, n_regions),
                np.tile(networks, n_subj),
                np.tile(regions, n_subj),
                metric.flatten()
            ]
            if long_data is None:
                long_data = cols
            else:
                long_data = np.r_[long_data, cols]
    df = DataFrame(long_data, columns=["task", "scan", "task_scan", "subject", "network", "region", "metric"])
    df.metric = df.metric.astype(np.float)
    return df


@pf.register_dataframe_method
def build_single_topo_map(df: DataFrame, template_name: TemplateName, space: Space):
    """
    :param df: must only contains two columns: regions and values
    :param template_name:
    :param space:
    :return:
    """
    mask, _, networks, regions, brain_axis = get_template(template_name, space)
    topo = np.zeros((1, mask.size))
    values = df.values
    for reg, pc in values:
        reg_index = np.argmax(regions == reg) + 1
        if reg_index == 0:
            print("0 reg_index in %s" % reg)
        topo[0, np.argwhere(mask == reg_index)] = pc
    return topo, brain_axis


def combine_topo_map(topo_maps):
    topos, brains = list(zip(*topo_maps))
    return np.concatenate(topos, axis=0), brains[0], cifti.Series(0, 1, len(topo_maps))
